from enum import IntEnum
import neuromeka.proto.common_msgs_pb2 as common_msgs
import neuromeka.proto.control_msgs_pb2 as control_msgs

from dataclasses import dataclass, field
from typing import List, Tuple


class StopCategory(IntEnum):
    CAT0 = common_msgs.IMMEDIATE_BRAKE
    CAT1 = common_msgs.SMOOTH_BRAKE
    CAT2 = common_msgs.SMOOTH_ONLY

class JointBaseType(IntEnum):
    ABSOLUTE = control_msgs.ABSOLUTE_JOINT
    RELATIVE = control_msgs.RELATIVE_JOINT

class TaskBaseType(IntEnum):
    ABSOLUTE = control_msgs.ABSOLUTE_TASK
    RELATIVE = control_msgs.RELATIVE_TASK
    TCP = control_msgs.TCP_TASK

class InterpolatorType(IntEnum):
    VELOCITY = 0
    TIME = 1


class CircularSettingType(IntEnum):
    POINT_SET = control_msgs.POINT_SET
    CENTER_AXIS = control_msgs.CENTER_AXIS


class BlendingType:
    NONE = control_msgs.BlendingType.NONE
    OVERRIDE = control_msgs.BlendingType.OVERRIDE
    DUPLICATE = control_msgs.BlendingType.DUPLICATE
    RADIUS = 4


class TrajCondType:
    STARTED = common_msgs.TRAJ_STARTED
    ACC_DONE = common_msgs.TRAJ_ACC_DONE
    CRZ_DONE = common_msgs.TRAJ_CRZ_DONE
    DEC_DONE = common_msgs.TRAJ_DEC_DONE


class TrajState:
    NONE = common_msgs.TrajState.TRAJ_NONE
    INIT = common_msgs.TRAJ_INIT
    CALC = common_msgs.TRAJ_CALC
    STAND_BY  = common_msgs.TRAJ_STAND_BY
    ACC = common_msgs.TRAJ_ACC
    CRUISE = common_msgs.TRAJ_CRUISE
    DEC = common_msgs.TRAJ_DEC
    CANCELLING = common_msgs.TRAJ_CANCELLING
    FINISHED = common_msgs.TRAJ_FINISHED
    ERROR = common_msgs.TRAJ_ERROR


class BlendingCondType:
    TIME = 0
    DIO = 1
    PROGRESS = 2
    ACCELERATION = 3
    CONSTSPEED = 4
    DECELERATION = 5
    RADIUS = 6
    EXPRESSION = 7


class CircularSettingType:
    POINT_SET = 0
    CENTER_AXIS = 1


class CircularMovingType:
    CONSTANT = 0
    RADIAL = 1
    SMOOTH = 2


class StopType:
    IMMEDIATE_BRAKE = common_msgs.IMMEDIATE_BRAKE
    SLOW_AND_BRAKE = common_msgs.SMOOTH_BRAKE
    SLOW = common_msgs.SMOOTH_ONLY


class PauseType:
    SMOOTH = common_msgs.SMOOTH_PAUSE
    IMMEDIATE = common_msgs.IMMEDIATE_PAUSE



class CollisionPolicyType:
    NONE = common_msgs.COLL_NO_DETECT
    NO_DETECT = common_msgs.COLL_NO_DETECT
    PAUSE = common_msgs.COLL_PAUSE
    RESUME_AFTER_SLEEP = common_msgs.COLL_RESUME_AFTER_SLEEP
    STOP = common_msgs.COLL_STOP


class ConditionType:
    CONST_CONT = control_msgs.MotionCondition.CONST_COND
    IO_CONT = control_msgs.MotionCondition.IO_COND
    VAR_COND = control_msgs.MotionCondition.VAR_COND


class ReactionType:
    NONE = 0  # control_data.MotionCondition.NONE_COND
    STOP = 1  # control_data.MotionCondition.STOP_COND
    PAUSE = 2  # control_data.MotionCondition.PAUSE_COND


@dataclass
class PostCondition:
    condition_type: ConditionType = ConditionType.CONST_CONT
    reaction_type: ReactionType = ReactionType.NONE
    const_cond: bool = True
    digital_inputs: List[Tuple[int, bool]] = field(default_factory=list)
    i_vars: List[Tuple[str, int]] = field(default_factory=list)
    f_vars: List[Tuple[str, float]] = field(default_factory=list)
    b_vars: List[Tuple[str, bool]] = field(default_factory=list)
    m_vars: List[Tuple[str, int]] = field(default_factory=list)
    j_vars: List[Tuple[str, List[float]]] = field(default_factory=list)
    t_vars: List[Tuple[str, List[float]]] = field(default_factory=list)


# Boundaries for Indy High-level Command
class Limits(IntEnum):
    LevelMin = 1
    LevelMax = 9
    JogLevelMin = 1
    JogLevelMax = 3
    JogVelLevelDefault = 2
    JogAccLevelDefault = 2

    JogVelRatioDefault = 15  # %
    JogAccRatioDefault = 100  # %
    VelRatioMax = 100  # %
    VelRatioMin = 1  # %
    AccRatioMax = 900  # %
    AccRatioMin = 1  # %

    JogVelRatioMin = 5  # %
    JogVelRatioMax = 25  # %
    VelAutoLevelValue = (VelRatioMax - JogVelRatioMax) / (LevelMax - JogLevelMax)  # %
    VelManualLevelValue = (JogVelRatioMax - JogVelRatioMin) / (JogLevelMax - JogLevelMin)  # %

    TaskDispVelValueDefault = 250  # 250mm/s
    TaskDispAccValueDefault = 250  # 250mm/s^2
    TaskDispVelValueMax = 1000  # mm/s
    TaskRotVelValueMax = 120  # deg/s

    ExternalMotorSpeedMax = 250  # mm/s : 3000rpm -> 50 rev/sec * 5 mm/rev -> 250 mm/s




class DetectType:
    DETECT = 0
    RETRIEVE = 1


class VisionServerType:
    INDYEYE = 0
    PICKIT = 1
    OMRON = 2


@dataclass
class VisionFrame:
    vision_server: str
    target_object: str


@dataclass
class VisionServer:
    name: str
    vision_server_type: VisionServerType
    ip: str
    port: str


@dataclass
class VisionResult:
    object_name: str
    detected: bool
    passed: bool
    frame: list


# @dataclass
# class CollisionConfig:
#     policy: CollisionPolicyType = CollisionPolicyType.PAUSE
#     sleep_time: int = 5


# @dataclass
# class CollisionTuning:
#     precision: TuningPrecision = TuningPrecision.MIDDLE
#     tuning_space: TuningSpace = TuningSpace.NONE
#     vel_level_max: int = 3



# @dataclass
# class Blend:
#     blending_type: BlendingType = BlendingType.NONE
#     blending_condition_type: BlendingCondType = BlendingCondType.RADIUS
#     conjunction: int = 0
#     async_sleep: bool = True
#
#     traj_radius: float = 0.0
#     time: int = -1
#     digital_outputs: List[Tuple[int, bool]] = field(default_factory=list)
#     digital_inputs: List[Tuple[int, bool]] = field(default_factory=list)
#     traj_progress: int = -1

class EtherCATStatus(IntEnum):
    OP_MODE_NO_MODE = 0x00
    OP_MODE_PROFILE_POSITION = 0x01
    OP_MODE_VELOCITY = 0x02
    OP_MODE_PROFILE_VELOCITY = 0x03
    OP_MODE_TORQUE_PROFILE = 0x04
    OP_MODE_HOMING = 0x06
    OP_MODE_INTERPOLATED_POSITION = 0x07
    OP_MODE_CYCLIC_SYNC_POSITION = 0x08
    OP_MODE_CYCLIC_SYNC_VELOCITY = 0x09
    OP_MODE_CYCLIC_SYNC_TORQUE = 0x0a

    @staticmethod
    def status2string(statusword):
        if (((statusword) & 0x004f) == 0x0000):  # x0xx 0000
            return "NOT_READY"
        elif (((statusword) & 0x004f) == 0x0040):  # x1xx 0000
            return "SWITCH_DISABLED"
        elif (((statusword) & 0x006f) == 0x0021):  # x01x 0001
            return "READY_SWITCH"
        elif (((statusword) & 0x006f) == 0x0023):  # x01x 0011
            return "SWITCHED_ON"
        elif (((statusword) & 0x006f) == 0x0027):  # x01x 0111
            return "OPERATION_ENABLED"
        elif (((statusword) & 0x006f) == 0x0007):  # x00x 0111
            return "QUICK_STOP"
        elif (((statusword) & 0x004f) == 0x000f):  # x0xx 1111
            return "FAULT_REACTION"
        elif (((statusword) & 0x004f) == 0x0008):  # x0xx 1000
            return "FAULT"
        else:
            return "UNKNOWN"

    @staticmethod
    def modeop2string(modeop):
        if modeop == 0x00:
            return "None"
        elif modeop == 0x01:
            return "PP"
        elif modeop == 0x03:
            return "PV"
        elif modeop == 0x04:
            return "TP"
        elif modeop == 0x06:
            return "Homing"
        elif modeop == 0x08:
            return "CSP"
        elif modeop == 0x09:
            return "CSV"
        elif modeop == 0x0a:
            return "CST"
        else:
            return "UNKNOWN"

    @staticmethod
    def error_code(mode_op, status_word):
        string_out = []
        if mode_op == EtherCATStatus.OP_MODE_PROFILE_POSITION:
            if (status_word & 0x2000):
                string_out.append("Following error")
            if (status_word & 0x1000):
                string_out.append("Set-point acknowledge")
            if (status_word & 0x0400):
                string_out.append("Target reached")

        elif mode_op == EtherCATStatus.OP_MODE_PROFILE_VELOCITY:
            if (status_word & 0x2000):
                string_out.append("Max slippage error")
            if (status_word & 0x1000):
                string_out.append("Speed")
            if (status_word & 0x0400):
                string_out.append("Target reached")

        elif mode_op == EtherCATStatus.OP_MODE_CYCLIC_SYNC_POSITION:
            if (status_word & 0x2000):
                string_out.append("Following error")
            if (status_word & 0x1000):
                string_out.append("Drive follows command value")

        elif mode_op == EtherCATStatus.OP_MODE_CYCLIC_SYNC_VELOCITY:
            if (status_word & 0x1000):
                string_out.append("Drive follows command value")

        elif mode_op == EtherCATStatus.OP_MODE_CYCLIC_SYNC_TORQUE:
            if (status_word & 0x1000):
                string_out.append("Drive follows command value")

        return string_out


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import common_msgs_pb2 as common__msgs__pb2
import device_msgs_pb2 as device__msgs__pb2


class DeviceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SetBrakes = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetBrakes',
                request_serializer=device__msgs__pb2.MotorList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetServoAll = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetServoAll',
                request_serializer=common__msgs__pb2.State.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetServo = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetServo',
                request_serializer=device__msgs__pb2.Servo.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetDI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetDI',
                request_serializer=device__msgs__pb2.DigitalList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetDO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetDO',
                request_serializer=device__msgs__pb2.DigitalList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetEndDI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetEndDI',
                request_serializer=device__msgs__pb2.EndtoolSignalList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetEndDO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetEndDO',
                request_serializer=device__msgs__pb2.EndtoolSignalList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetAI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetAI',
                request_serializer=device__msgs__pb2.AnalogList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetAO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetAO',
                request_serializer=device__msgs__pb2.AnalogList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetEndAI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetEndAI',
                request_serializer=device__msgs__pb2.AnalogList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.SetEndAO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetEndAO',
                request_serializer=device__msgs__pb2.AnalogList.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.GetDI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetDI',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.DigitalList.FromString,
                )
        self.GetDO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetDO',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.DigitalList.FromString,
                )
        self.GetEndDI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetEndDI',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.EndtoolSignalList.FromString,
                )
        self.GetEndDO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetEndDO',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.EndtoolSignalList.FromString,
                )
        self.GetAI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetAI',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.AnalogList.FromString,
                )
        self.GetAO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetAO',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.AnalogList.FromString,
                )
        self.GetEndAI = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetEndAI',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.AnalogList.FromString,
                )
        self.GetEndAO = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetEndAO',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.AnalogList.FromString,
                )
        self.GetEL5001 = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetEL5001',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.GetEL5001Data.FromString,
                )
        self.GetEL5101 = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetEL5101',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.GetEL5101Data.FromString,
                )
        self.GetDeviceInfo = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetDeviceInfo',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.DeviceInfo.FromString,
                )
        self.GetBrakeControlStyle = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetBrakeControlStyle',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.BrakeControlStyle.FromString,
                )
        self.SetFTSensorType = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/SetFTSensorType',
                request_serializer=device__msgs__pb2.FTSensorType.SerializeToString,
                response_deserializer=common__msgs__pb2.Response.FromString,
                )
        self.GetFTSensorType = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetFTSensorType',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.FTSensorType.FromString,
                )
        self.GetFTSensorData = channel.unary_unary(
                '/Nrmk.IndyFramework.Device/GetFTSensorData',
                request_serializer=common__msgs__pb2.Empty.SerializeToString,
                response_deserializer=device__msgs__pb2.FTSensorData.FromString,
                )


class DeviceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SetBrakes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetServoAll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetServo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetDI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetEndDI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetEndDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetAI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetEndAI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetEndAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEndDI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEndDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEndAI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEndAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEL5001(self, request, context):
        """beckhoff_conv
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEL5101(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDeviceInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBrakeControlStyle(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetFTSensorType(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFTSensorType(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFTSensorData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DeviceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SetBrakes': grpc.unary_unary_rpc_method_handler(
                    servicer.SetBrakes,
                    request_deserializer=device__msgs__pb2.MotorList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetServoAll': grpc.unary_unary_rpc_method_handler(
                    servicer.SetServoAll,
                    request_deserializer=common__msgs__pb2.State.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetServo': grpc.unary_unary_rpc_method_handler(
                    servicer.SetServo,
                    request_deserializer=device__msgs__pb2.Servo.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetDI': grpc.unary_unary_rpc_method_handler(
                    servicer.SetDI,
                    request_deserializer=device__msgs__pb2.DigitalList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetDO': grpc.unary_unary_rpc_method_handler(
                    servicer.SetDO,
                    request_deserializer=device__msgs__pb2.DigitalList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetEndDI': grpc.unary_unary_rpc_method_handler(
                    servicer.SetEndDI,
                    request_deserializer=device__msgs__pb2.EndtoolSignalList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetEndDO': grpc.unary_unary_rpc_method_handler(
                    servicer.SetEndDO,
                    request_deserializer=device__msgs__pb2.EndtoolSignalList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetAI': grpc.unary_unary_rpc_method_handler(
                    servicer.SetAI,
                    request_deserializer=device__msgs__pb2.AnalogList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetAO': grpc.unary_unary_rpc_method_handler(
                    servicer.SetAO,
                    request_deserializer=device__msgs__pb2.AnalogList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetEndAI': grpc.unary_unary_rpc_method_handler(
                    servicer.SetEndAI,
                    request_deserializer=device__msgs__pb2.AnalogList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'SetEndAO': grpc.unary_unary_rpc_method_handler(
                    servicer.SetEndAO,
                    request_deserializer=device__msgs__pb2.AnalogList.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'GetDI': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDI,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.DigitalList.SerializeToString,
            ),
            'GetDO': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDO,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.DigitalList.SerializeToString,
            ),
            'GetEndDI': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEndDI,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.EndtoolSignalList.SerializeToString,
            ),
            'GetEndDO': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEndDO,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.EndtoolSignalList.SerializeToString,
            ),
            'GetAI': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAI,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.AnalogList.SerializeToString,
            ),
            'GetAO': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAO,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.AnalogList.SerializeToString,
            ),
            'GetEndAI': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEndAI,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.AnalogList.SerializeToString,
            ),
            'GetEndAO': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEndAO,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.AnalogList.SerializeToString,
            ),
            'GetEL5001': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEL5001,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.GetEL5001Data.SerializeToString,
            ),
            'GetEL5101': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEL5101,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.GetEL5101Data.SerializeToString,
            ),
            'GetDeviceInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDeviceInfo,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.DeviceInfo.SerializeToString,
            ),
            'GetBrakeControlStyle': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBrakeControlStyle,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.BrakeControlStyle.SerializeToString,
            ),
            'SetFTSensorType': grpc.unary_unary_rpc_method_handler(
                    servicer.SetFTSensorType,
                    request_deserializer=device__msgs__pb2.FTSensorType.FromString,
                    response_serializer=common__msgs__pb2.Response.SerializeToString,
            ),
            'GetFTSensorType': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFTSensorType,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.FTSensorType.SerializeToString,
            ),
            'GetFTSensorData': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFTSensorData,
                    request_deserializer=common__msgs__pb2.Empty.FromString,
                    response_serializer=device__msgs__pb2.FTSensorData.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Nrmk.IndyFramework.Device', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Device(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SetBrakes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetBrakes',
            device__msgs__pb2.MotorList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetServoAll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetServoAll',
            common__msgs__pb2.State.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetServo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetServo',
            device__msgs__pb2.Servo.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetDI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetDI',
            device__msgs__pb2.DigitalList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetDO',
            device__msgs__pb2.DigitalList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetEndDI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetEndDI',
            device__msgs__pb2.EndtoolSignalList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetEndDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetEndDO',
            device__msgs__pb2.EndtoolSignalList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetAI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetAI',
            device__msgs__pb2.AnalogList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetAO',
            device__msgs__pb2.AnalogList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetEndAI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetEndAI',
            device__msgs__pb2.AnalogList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetEndAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetEndAO',
            device__msgs__pb2.AnalogList.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetDI',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.DigitalList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetDO',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.DigitalList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEndDI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetEndDI',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.EndtoolSignalList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEndDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetEndDO',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.EndtoolSignalList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetAI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetAI',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.AnalogList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetAO',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.AnalogList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEndAI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetEndAI',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.AnalogList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEndAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetEndAO',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.AnalogList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEL5001(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetEL5001',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.GetEL5001Data.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEL5101(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetEL5101',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.GetEL5101Data.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDeviceInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetDeviceInfo',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.DeviceInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetBrakeControlStyle(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetBrakeControlStyle',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.BrakeControlStyle.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetFTSensorType(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/SetFTSensorType',
            device__msgs__pb2.FTSensorType.SerializeToString,
            common__msgs__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFTSensorType(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetFTSensorType',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.FTSensorType.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFTSensorData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Nrmk.IndyFramework.Device/GetFTSensorData',
            common__msgs__pb2.Empty.SerializeToString,
            device__msgs__pb2.FTSensorData.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import psutil
import sys, json, grpc, time
import paramiko
import csv
import math

from matplotlib.pyplot import figure
import matplotlib.pyplot as plt


PI = math.pi
RAD2DEG = 180/PI
DEG2RAD = PI/180

class PlotTool:
    def __init__(self):
        self.record = {
            "t": [], "cycle_time": [], "period": [],
            **{f"q{i}": [] for i in range(1, 7)},
            **{f"qdot{i}": [] for i in range(1, 7)},
            **{f"qddot{i}": [] for i in range(1, 7)},
            **{f"tau{i}": [] for i in range(1, 7)},
            **{f"tauact{i}": [] for i in range(1, 7)},
            **{f"qd{i}": [] for i in range(1, 7)},
            **{f"qdotd{i}": [] for i in range(1, 7)},
            **{f"qddotd{i}": [] for i in range(1, 7)},
            **{f"p{i}": [] for i in range(1, 7)},
            **{f"pd{i}": [] for i in range(1, 7)},
            **{f"qe{i}": [] for i in range(1, 7)},
            **{f"pe{i}": [] for i in range(1, 7)},
            **{f"fe{i}": [] for i in range(1, 7)},
            **{f"fdes{i}": [] for i in range(1, 7)},
            **{f"fact{i}": [] for i in range(1, 7)}
        }

    def load_data_6dof(self, file_name, step_ip):

        # Saved folder paramiko
        file_to_directory = '/home/user/release/IndyDeployment/RTLog'
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(step_ip, username='root', password='root')

        # open the remote file using `sftp`
        sftp = ssh.open_sftp()
        remote_file = sftp.open(file_to_directory+'/'+file_name)

        # read the contents of the file using `csv` library
        reader = csv.reader(remote_file, delimiter=",")
            

        for line in reader:
            # Appending values to lists
            self.record["t"].append(float(line[0]))
            self.record["cycle_time"].append(float(line[1]))
            self.record["period"].append(float(line[2]))

            for i in range(1, 7):
                self.record[f"q{i}"].append(float(line[2 + i]))
                self.record[f"qdot{i}"].append(float(line[8 + i]))
                self.record[f"qddot{i}"].append(float(line[14 + i]))
                self.record[f"tau{i}"].append(float(line[20 + i]))
                self.record[f"tauact{i}"].append(float(line[26 + i]))
                self.record[f"qd{i}"].append(float(line[32 + i]))
                self.record[f"qdotd{i}"].append(float(line[38 + i]))
                self.record[f"qddotd{i}"].append(float(line[44 + i]))
                self.record[f"p{i}"].append(float(line[50 + i]))
                self.record[f"pd{i}"].append(float(line[56 + i]))
                self.record[f"fdes{i}"].append(float(line[62 + i]))
                self.record[f"fact{i}"].append(float(line[68 + i]))

                # Compute Joint and Task Position errors
                self.record[f"qe{i}"].append(self.record[f"qd{i}"][-1] - self.record[f"q{i}"][-1])
                self.record[f"pe{i}"].append(self.record[f"pd{i}"][-1] - self.record[f"p{i}"][-1])
                self.record[f"fe{i}"].append(self.record[f"fdes{i}"][-1] - self.record[f"fact{i}"][-1])



    def load_data_7dof(self, step_ip, file_name):
        # Saved folder paramiko
        file_to_directory = '/home/user/release/IndyDeployment/RTLog'
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(step_ip, username='root', password='root')

        # open the remote file using `sftp`
        sftp = ssh.open_sftp()
        remote_file = sftp.open(file_to_directory+'/'+file_name)

        # read the contents of the file using `csv` library
        reader = csv.reader(remote_file, delimiter=",")

        # Initializing lists
        for line in reader:
            # Appending values to lists
            self.record["t"].append(float(line[0]))
            self.record["cycle_time"].append(float(line[1]))
            self.record["period"].append(float(line[2]))

            for i in range(1, 8):
                self.record[f"q{i}"].append(float(line[2 + i]))
                self.record[f"qdot{i}"].append(float(line[9 + i]))
                self.record[f"qddot{i}"].append(float(line[16 + i]))
                self.record[f"tau{i}"].append(float(line[23 + i]))
                self.record[f"tauact{i}"].append(float(line[30 + i]))
                self.record[f"qd{i}"].append(float(line[37 + i]))
                self.record[f"qdotd{i}"].append(float(line[44 + i]))
                self.record[f"qddotd{i}"].append(float(line[51 + i]))
                self.record[f"p{i}"].append(float(line[58 + i]))
                self.record[f"pd{i}"].append(float(line[65 + i]))

                # Compute Joint and Task Position errors
                self.record[f"qe{i}"].append(self.record[f"qd{i}"][-1] - self.record[f"q{i}"][-1])
                self.record[f"pe{i}"].append(self.record[f"pd{i}"][-1] - self.record[f"p{i}"][-1])


    def get_raw_data(self):
        return self.record

    ################################
    # Plot various data
    ################################
    def plot_robot_path(self):
        plt.figure(figsize=(8, 6), dpi=80)
        plt.plot(self.record['p1'], self.record['p2'], 'b', label='Actual Path')
        plt.xlabel('x Position (m)')
        plt.ylabel('y Position (m)')
        plt.title('Robot Path in x-y Plane')
        plt.legend()
        plt.grid(True)
        plt.show()

    def plot_robot_and_desired_path(self):
        plt.figure(figsize=(8, 6), dpi=80)
        plt.plot(self.record['p1'], self.record['p2'], 'b', label='Actual Path')
        plt.plot(self.record['pd1'], self.record['pd2'], 'r--', label='Desired Path')
        plt.xlabel('x Position (m)')
        plt.ylabel('y Position (m)')
        plt.title('Robot and Desired Path in x-y Plane')
        plt.legend()
        plt.grid(True)
        plt.show()

    def plot_robot_and_desired_force(self):
        plt.figure(figsize=(8, 6), dpi=80)
        plt.plot(self.record['t'], self.record['fact1'], 'b', label='Actual Path')
        plt.plot(self.record['t'], self.record['fdes1'], 'r--', label='Desired Path')
        plt.xlabel('x Time (s)')
        plt.ylabel('y Force (Nm)')
        plt.title('Force tracking plot')
        plt.legend()
        plt.grid(True)
        plt.show()

    ## Joint plots
    def plot_joint_pos(self, unit='rad'):
        fig, axes = plt.subplots(6, 1, figsize=(10, 15), dpi=80)
        for i, ax in enumerate(axes, start=1):
            if unit=='deg':
                q = [x*RAD2DEG for x in self.record[f'q{i}']]
                qd = [x*RAD2DEG for x in self.record[f'qd{i}']]                
                ax.set_ylabel(f'q{i}: joint position (deg)')
            elif unit=='rad':
                q = self.record[f'q{i}']
                qd = self.record[f'qd{i}']                
                ax.set_ylabel(f'q{i}: joint position (rad)')
            ax.plot(self.record['t'], q, 'b', self.record['t'], qd, 'r--')
            ax.set_xlabel('Time [sec]')            
            ax.grid(True)            
            max_qd = max(qd)
            min_qd = min(qd)            
            ax.text(self.record['t'][qd.index(max_qd)], max_qd, f'Max qd: {max_qd:.2f}', color='red')
            ax.text(self.record['t'][qd.index(min_qd)], min_qd, f'Min qd: {min_qd:.2f}', color='red', verticalalignment='top')
        fig.suptitle('Joint Positions')
        plt.tight_layout()
        plt.show()        
        


    def plot_joint_vel(self, unit='rad'):
        fig, axes = plt.subplots(6, 1, figsize=(10, 15), dpi=80)
        for i, ax in enumerate(axes, start=1):
            if unit=='deg':
                qdot = [x*RAD2DEG for x in self.record[f'qdot{i}']]
                qdotd = [x*RAD2DEG for x in self.record[f'qdotd{i}']]                
                ax.set_ylabel(f'qdot{i}: joint velocity (deg/s)')
            elif unit=='rad':
                qdot = self.record[f'qdot{i}']
                qdotd = self.record[f'qdotd{i}']                
                ax.set_ylabel(f'qdot{i}: joint velocity (rad/s)')
            ax.plot(self.record['t'], qdot, 'b', self.record['t'], qdotd, 'r--')
            ax.set_xlabel('Time')            
            ax.grid(True)
            max_qdotd = max(qdotd)
            min_qdotd = min(qdotd)            
            ax.text(self.record['t'][qdotd.index(max_qdotd)], max_qdotd, f'Max qdotd: {max_qdotd:.2f}', color='red')
            ax.text(self.record['t'][qdotd.index(min_qdotd)], min_qdotd, f'Min qdotd: {min_qdotd:.2f}', color='red', verticalalignment='top')
        fig.suptitle('Joint Velocity')
        plt.tight_layout()
        plt.show()

    def plot_joint_acc(self, unit='rad'):
        fig, axes = plt.subplots(6, 1, figsize=(10, 15), dpi=80)
        for i, ax in enumerate(axes, start=1):
            if unit=='deg':
                qddot = [x*RAD2DEG for x in self.record[f'qddot{i}']]
                qddotd = [x*RAD2DEG for x in self.record[f'qddotd{i}']]                
                ax.set_ylabel(f'qddot{i}: joint acceleration (deg/s2)')
            elif unit=='rad':
                qddot = self.record[f'qddot{i}']
                qddotd = self.record[f'qddotd{i}']
                ax.set_ylabel(f'qddot{i}: joint acceleration (rad/s2)')
            ax.plot(self.record['t'], qddot, 'b', self.record['t'], qddotd, 'r--')
            ax.set_xlabel('Time')            
            ax.grid(True)
            max_qddotd = max(qddotd)
            min_qddotd = min(qddotd)            
            ax.text(self.record['t'][qddotd.index(max_qddotd)], max_qddotd, f'Max qddotd: {max_qddotd:.2f}', color='red')
            ax.text(self.record['t'][qddotd.index(min_qddotd)], min_qddotd, f'Min qddotd: {min_qddotd:.2f}', color='red', verticalalignment='top')
        fig.suptitle('Joint Acceleration')
        plt.tight_layout()
        plt.show()

    def plot_jerror(self):
        fig, (ax1) = plt.subplots(1, 1, figsize=(10, 5), dpi=80)
        ax1.plot(self.record['t'], self.record['qe1'], self.record['t'], self.record['qe2'], self.record['t'], self.record['qe3'], \
                 self.record['t'], self.record['qe4'], self.record['t'], self.record['qe5'], self.record['t'], self.record['qe6'])
        ax1.set_xlabel('Time')
        ax1.set_ylabel('Error: joint position (rad)')
        ax1.grid(True)
        fig.suptitle('Joint Positino Error')
        plt.tight_layout()
        plt.show()

    def plot_joint_torque(self):
        fig, axes = plt.subplots(6, 1, figsize=(10, 15), dpi=80)
        for i, ax in enumerate(axes, start=1):
            ax.plot(self.record['t'], self.record[f'tau{i}'], 'b', self.record['t'], self.record[f'tauact{i}'], 'r--')
            ax.set_xlabel('Time')
            ax.set_ylabel(f'tau{i}: joint torque')
            ax.grid(True)
        fig.suptitle('Joint Torque')
        plt.tight_layout()
        plt.show()


def plot_task(record):
    def plot_task_pos():
        fig, axes = plt.subplots(6, 1, figsize=(10, 15), dpi=80)
        for i, ax in enumerate(axes, start=1):
            ax.plot(record['t'], record[f'p{i}'], 'b', record['t'], record[f'pd{i}'], 'r--')
            ax.set_xlabel('Time')
            ax.set_ylabel(f'p{i}: Task position (m/rad)')
            ax.grid(True)
        fig.suptitle('Joint Positions')
        plt.tight_layout()
        plt.show()
    def plot_terror():
        fig, (ax1) = plt.subplots(1, 1, figsize=(10, 5), dpi=80)
        ax1.plot(record['t'], record['pe1'], record['t'], record['pe2'], record['t'], record['pe3'], \
                 record['t'], record['pe4'], record['t'], record['pe5'], record['t'], record['pe6'])
        ax1.set_xlabel('Time')
        ax1.set_ylabel('Error: joint position (rad)')
        fig.suptitle('Joint Positions')
        plt.tight_layout()
        ax1.grid(True)

    plot_task_pos()
    plot_terror()
    



from neuromeka.proto import *
from neuromeka.common import *
from neuromeka.enums import *


import grpc
from google.protobuf import json_format
from google.protobuf.json_format import ParseDict

CONTROL_SOCKET_PORT = [20001, 30001]
DEVICE_SOCKET_PORT = [20002, 30002]
CONFIG_SOCKET_PORT = [20003, 30003]
RTDE_SOCKET_PORT = [20004, 30004]


class IndyDCP3:
    def __init__(self, robot_ip='127.0.0.1', index=0):
        if index not in [0, 1]:
            raise ValueError("Index must be 0 or 1")

        self.control_channel = grpc.insecure_channel('{}:{}'.format(robot_ip, CONTROL_SOCKET_PORT[index]))
        self.device_channel = grpc.insecure_channel('{}:{}'.format(robot_ip, DEVICE_SOCKET_PORT[index]))
        self.config_channel = grpc.insecure_channel('{}:{}'.format(robot_ip, CONFIG_SOCKET_PORT[index]))
        self.rtde_channel = grpc.insecure_channel('{}:{}'.format(robot_ip, RTDE_SOCKET_PORT[index]))

        self.control = ControlStub(self.control_channel)
        self.device = DeviceStub(self.device_channel)
        self.config = ConfigStub(self.config_channel)
        self.rtde = RTDataExchangeStub(self.rtde_channel)

    def __del__(self):
        if self.control_channel is not None:
            self.control_channel.close()
        if self.device_channel is not None:
            self.device_channel.close()
        if self.config_channel is not None:
            self.config_channel.close()
        if self.rtde_channel is not None:
            self.rtde_channel.close()

    def __to_digital_request_list__(self, digital_signal_list) -> list:
        request_list = []
        if digital_signal_list is not None:
            for signal in digital_signal_list:
                request_list.append(device_msgs.DigitalSignal(address=signal['address'], state=signal['state']))
        return request_list

    def __to_analog_request_list__(self, analog_signal_list) -> list:
        request_list = []
        if analog_signal_list is not None:
            for signal in analog_signal_list:
                request_list.append(device_msgs.AnalogSignal(address=signal['address'], voltage=signal['voltage']))
        return request_list

    ############################
    # IndyDCP3 API protocols
    ############################
    def get_motion_data(self):
        """
        Motion Data:
            traj_state   -> TrajState
            traj_progress   -> int32
            is_in_motion  -> bool
            is_target_reached  -> bool
            is_pausing  -> bool
            is_stopping  -> bool
            has_motion  -> bool
            speed_ratio  -> int32
            motion_id  -> int32
            remain_distance  -> float
            motion_queue_size  -> uint32
            cur_traj_progress  -> int32
        """
        response = self.rtde.GetMotionData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_control_data(self):
        """
        Control Data:
            running_hours   -> uint32
            running_mins   -> uint32
            running_secs  -> uint32
            op_state  -> OpState
            sim_mode  -> bool
            q  -> float[6]
            qdot  -> float[6]
            p  -> float[6]
            pdot  -> float[6]
            ref_frame  -> float[6]
            tool_frame  -> float[6]
            response  -> Response
        """
        response = self.rtde.GetControlData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_control_state(self):
        """
        Control Data:
            q  -> float[]
            qdot  -> float[]
            qddot  -> float[]
            qdes  -> float[]
            qdotdes  -> float[]
            qddotdes  -> float[]
            p  -> float[]
            pdot  -> float[]
            pddot  -> float[]
            pdes  -> float[]
            pdotdes  -> float[]
            pddotdes  -> float[]
            tau  -> float[]
            tau_act  -> float[]
            tau_ext  -> float[]
        """
        response = self.rtde.GetControlState(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_servo_data(self):
        """
        Servo Data:
            status_codes   -> string[]
            temperatures   -> float[]
            voltages  -> float[]
            currents  -> float[]
            servo_actives  -> bool[]
            brake_actives  -> bool[]
        """
        response = self.rtde.GetServoData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_violation_data(self):
        """
        Violation Data:
            violation_code   -> uint64
            j_index   -> uint32
            i_args  -> int32[]
            f_args  -> float[]
            violation_str  -> string
        """
        response = self.rtde.GetViolationData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_program_data(self):
        """
        Program Data:
            program_state   -> ProgramState
            cmd_id   -> int32
            sub_cmd_id  -> int32
            running_hours  -> int32
            running_mins  -> int32
            running_secs  -> int32
            program_name  -> string
            program_alarm  -> string
            program_annotation  -> string
        """
        response = self.rtde.GetProgramData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    ############################
    # IO board and Endtool port interfaces
    ############################
    def get_di(self):
        """
        address = uint32
        state = DigitalState
        """
        response = self.device.GetDI(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_do(self):
        """
        signals = index
        address = uint32
        state = DigitalState
        """
        response = self.device.GetDO(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_do(self, do_signal_list: list):
        """
        do_list = [(int_addr1, True/False), (int_addr1, True/False), ...]
        """
        response = self.device.SetDO(device_msgs.DigitalList(
            signals=do_signal_list,
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_ai(self) -> list:
        """
        address = uint32
        voltage = int32
        """
        response = self.device.GetAI(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_ao(self) -> list:
        """
        address = uint32
        voltage = int32
        """
        response = self.device.GetAO(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_ao(self, ao_signal_list: list):
        response = self.device.SetAO(device_msgs.AnalogList(
            signals=ao_signal_list,
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_endtool_di(self) -> list:
        """
        state = EndtoolState
        port = char value [A,B,C]
        """
        response = self.device.GetEndDI(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_endtool_do(self) -> list:
        """
        state = EndtoolState
        port = char value [A,B,C]
        """
        response = self.device.GetEndDO(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_endtool_do(self, end_do_signal_list: list):
        response = self.device.SetEndDO(device_msgs.EndtoolSignalList(
            signals=end_do_signal_list,
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_endtool_ai(self) -> list:
        """
        address = uint32
        voltage = int32
        """
        response = self.device.GetEndAI(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_endtool_ao(self) -> list:
        """
        address = uint32
        voltage = int32
        """
        response = self.device.GetEndAO(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_endtool_ao(self, end_ao_signal_list: list):
        response = self.device.SetEndAO(device_msgs.AnalogList(
            signals=end_ao_signal_list,
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_device_info(self):
        """
        Device Info:
            num_joints   -> uint32
            robot_serial   -> string
            io_board_fw_ver  -> string
            core_board_fw_vers  -> string[6]
            endtool_board_fw_ver  -> string
            endtool_port_type  -> EndToolPortType
        """
        response = self.device.GetDeviceInfo(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_ft_sensor_data(self):
        """
        Device Info:
            num_joints   -> uint32
            robot_serial   -> string
            io_board_fw_ver  -> string
            core_board_fw_vers  -> string[6]
            endtool_board_fw_ver  -> string
            endtool_port_type  -> EndToolPortType
            response  -> {code: int64, msg: string}
        """
        response = self.device.GetFTSensorData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    ############################
    # Motion Control (Move commands)
    ############################
    def stop_motion(self, stop_category=StopCategory.CAT2) -> dict:
        """
         stop motion element:
            stop_category -> StopCategory
                CAT0  = 0
                CAT1  = 1
                CAT2  = 2
        """
        response = self.control.StopMotion(common_msgs.StopCat(category=stop_category))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movej(self, jtarget,
              blending_type=BlendingType.NONE,
              base_type=JointBaseType.ABSOLUTE,
              blending_radius=0.0,
              vel_ratio=Limits.JogVelRatioDefault,
              acc_ratio=Limits.JogAccRatioDefault,
              post_condition=PostCondition(),
              teaching_mode=False) -> dict:
        """
         Joint Move:
            blending_type -> BlendingType.Type
                NONE
                OVERRIDE
                DUPLICATE
            base_type -> JointBaseType
                ABSOLUTE
                RELATIVE
            vel_ratio (0-100) -> int
            acc_ratio (0-100) -> int
            post_condition -> PostCondition
            teaching_mode -> bool

        """
        jtarget = control_msgs.TargetJ(j_start=[], j_target=list(jtarget), base_type=base_type)
        blending = control_msgs.BlendingType(type=blending_type, blending_radius=blending_radius)
        post_cond = control_msgs.MotionCondition()
        if post_condition is not None:
            post_cond = control_msgs.MotionCondition(
                type_cond=post_condition.condition_type,
                type_react=post_condition.reaction_type,
                const_cond=post_condition.const_cond,
                io_cond=control_msgs.IOCondition(
                    di=self.__to_digital_request_list__(
                        [{'address': di[0], 'state': di[1]} for di in post_condition.digital_inputs]),
                    # di=self.__to_digital_request_list__(post_condition.digital_inputs),
                    # end_di=self.__to_digital_request_list__(post_condition['enddi_condition']),
                ),
            )

        response = self.control.MoveJ(control_msgs.MoveJReq(
            target=jtarget,
            blending=blending,
            vel_ratio=vel_ratio, acc_ratio=acc_ratio,
            post_condition=post_cond,
            teaching_mode=teaching_mode
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movej_time(self, jtarget,
                   blending_type=BlendingType.NONE,
                   base_type=JointBaseType.ABSOLUTE,
                   blending_radius=0.0,
                   move_time=2.0,
                   post_condition=PostCondition()) -> dict:
        """
        jtarget = [deg, deg, deg, deg, deg, deg]
        move_time = seconds
        """
        jtarget = control_msgs.TargetJ(j_start=[], j_target=list(jtarget), base_type=base_type)
        blending = control_msgs.BlendingType(type=blending_type, blending_radius=blending_radius)
        post_cond = control_msgs.MotionCondition()
        if post_condition is not None:
            post_cond = control_msgs.MotionCondition(
                type_cond=post_condition.condition_type,
                type_react=post_condition.reaction_type,
                const_cond=post_condition.const_cond,
                io_cond=control_msgs.IOCondition(
                    di=self.__to_digital_request_list__(
                        [{'address': di[0], 'state': di[1]} for di in post_condition.digital_inputs]),
                    # di=self.__to_digital_request_list__(post_condition.digital_inputs),
                    # end_di=self.__to_digital_request_list__(post_condition['enddi_condition']),
                ),
            )

        response = self.control.MoveJT(control_msgs.MoveJTReq(
            target=jtarget,
            blending=blending,
            time=move_time,
            post_condition=post_cond
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movel(self, ttarget,
              blending_type=BlendingType.NONE,
              base_type=TaskBaseType.ABSOLUTE,
              blending_radius=0.0,
              vel_ratio=Limits.JogVelRatioDefault,
              acc_ratio=Limits.JogAccRatioDefault,
              post_condition=PostCondition(),
              teaching_mode=False) -> dict:
        """
        tstart = [mm, mm, mm, deg, deg, deg]
        ttarget = [mm, mm, mm, deg, deg, deg]

            base_tye -> TaskBaseType
                ABSOLUTE
                RELATIVE
                TCP
        """
        ptarget = control_msgs.TargetP(t_start=[], t_target=list(ttarget), base_type=base_type)
        blending = control_msgs.BlendingType(type=blending_type, blending_radius=blending_radius)
        post_cond = control_msgs.MotionCondition()
        if post_condition is not None:
            post_cond = control_msgs.MotionCondition(
                type_cond=post_condition.condition_type,
                type_react=post_condition.reaction_type,
                const_cond=post_condition.const_cond,
                io_cond=control_msgs.IOCondition(
                    di=self.__to_digital_request_list__(
                        [{'address': di[0], 'state': di[1]} for di in post_condition.digital_inputs]),
                    # di=self.__to_digital_request_list__(post_condition.digital_inputs),
                    # end_di=self.__to_digital_request_list__(post_condition['enddi_condition']),
                ),
            )

        response = self.control.MoveL(control_msgs.MoveLReq(
            target=ptarget,
            blending=blending,
            vel_ratio=vel_ratio, acc_ratio=acc_ratio,
            post_condition=post_cond,
            teaching_mode=teaching_mode
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movec(self, tpos0, tpos1,
              blending_type=BlendingType.NONE,
              base_type=TaskBaseType.ABSOLUTE,
              angle=90.0,
              setting_type=CircularSettingType.POINT_SET,
              move_type=control_msgs.CONSTANT,
              blending_radius=0.0,
              vel_ratio=Limits.JogVelRatioDefault,
              acc_ratio=Limits.JogAccRatioDefault,
              post_condition=PostCondition(),
              teaching_mode=False) -> dict:
        """
        tstart = [mm, mm, mm, deg, deg, deg]
        ttarget = [mm, mm, mm, deg, deg, deg]
         Recover from violation
        """

        ctarget = control_msgs.TargetC(t_start=[], t_pos0=list(tpos0), t_pos1=list(tpos1),
                                       base_type=base_type)
        blending = control_msgs.BlendingType(type=blending_type, blending_radius=blending_radius)
        post_cond = control_msgs.MotionCondition()
        if post_condition is not None:
            post_cond = control_msgs.MotionCondition(
                type_cond=post_condition.condition_type,
                type_react=post_condition.reaction_type,
                const_cond=post_condition.const_cond,
                io_cond=control_msgs.IOCondition(
                    di=self.__to_digital_request_list__(
                        [{'address': di[0], 'state': di[1]} for di in post_condition.digital_inputs]),
                    # di=self.__to_digital_request_list__(post_condition.digital_inputs),
                    # end_di=self.__to_digital_request_list__(post_condition['enddi_condition']),
                ),
            )

        response = self.control.MoveC(control_msgs.MoveCReq(
            target=ctarget,
            blending=blending,
            angle=angle,
            setting_type=setting_type,
            move_type=move_type,
            vel_ratio=vel_ratio, acc_ratio=acc_ratio,
            post_condition=post_cond,
            teaching_mode=teaching_mode
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    ############################
    # Motion Control (Teaching mode)
    ############################
    def move_home(self):
        home_pos = self.get_home_pos()
        self.movej(home_pos,
                  blending_type=NO_BLENDING,
                  base_type=JointBaseType.ABSOLUTE,
                  blending_radius=0.0,
                  vel_ratio=Limits.JogVelRatioDefault,
                  acc_ratio=Limits.JogAccRatioDefault,
                  post_condition=PostCondition(),
                  teaching_mode=False)



    ############################
    # Motion Control (Teleoperation)
    ############################
    def start_teleop(self, method):
        """
        Start tele op
        method:
            TELE_TASK_ABSOLUTE = 0
            TELE_TASK_RELATIVE = 1
            TELE_JOINT_ABSOLUTE = 10
            TELE_JOINT_RELATIVE = 11
        """
        response = self.control.StartTeleOp(control_msgs.TeleOpState(mode=control_msgs.TeleMode.TELE_RAW, method=method))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def stop_teleop(self):
        """
        Stop tele op
        """
        response = self.control.StopTeleOp(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movetelej_abs(self, jpos, vel_ratio=1.0, acc_ratio=1.0):
        """
        Joint Teleoperation - Absolute
        jpos = [deg, deg, deg, deg, deg, deg]
        """
        response = self.control.MoveTeleJ(control_msgs.MoveTeleJReq(jpos=jpos, vel_ratio=vel_ratio, acc_ratio=acc_ratio,
                                                                    method=control_msgs.TELE_JOINT_ABSOLUTE))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movetelej_rel(self, jpos, vel_ratio=1.0, acc_ratio=1.0):
        """
        Joint Teleoperation - Relative
        jpos = [deg, deg, deg, deg, deg, deg]
        """
        response = self.control.MoveTeleJ(control_msgs.MoveTeleJReq(jpos=jpos, vel_ratio=vel_ratio, acc_ratio=acc_ratio,
                                                                    method=control_msgs.TELE_JOINT_RELATIVE))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movetelel_abs(self, tpos, vel_ratio=1.0, acc_ratio=1.0):
        """
        Task Teleoperation - Absolute
        jpos = [mm, mm, mm, deg, deg, deg]
        """
        response = self.control.MoveTeleL(control_msgs.MoveTeleLReq(tpos=tpos, vel_ratio=vel_ratio, acc_ratio=acc_ratio,
                                      method=control_msgs.TELE_TASK_ABSOLUTE))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def movetelel_rel(self, tpos, vel_ratio=1.0, acc_ratio=1.0):
        """
        Task Teleoperation - Relative
        jpos = [mm, mm, mm, deg, deg, deg]
        """
        response = self.control.MoveTeleL(control_msgs.MoveTeleLReq(tpos=tpos, vel_ratio=vel_ratio, acc_ratio=acc_ratio,
                                      method=control_msgs.TELE_TASK_RELATIVE))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    ############################
    # Control - Additional
    ############################
    def inverse_kin(self, tpos, init_jpos) -> dict:
        """
        :param tpos:
        :param init_jpos:
        :return:
            'jpos': []
        """
        response = self.control.InverseKinematics(control_msgs.InverseKinematicsReq(
            tpos=list(tpos),
            init_jpos=list(init_jpos)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_direct_teaching(self, enable=True) -> dict:
        """
         enable = True | False
        """
        response = self.control.SetDirectTeaching(common_msgs.State(enable=enable))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_simulation_mode(self, enable=True) -> dict:
        """
         Set simulation mode = True | False
        """
        response = self.control.SetSimulationMode(common_msgs.State(enable=enable))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def recover(self) -> dict:
        """
         Recover from violation
        """
        response = self.control.Recover(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_manual_recovery(self, enable=True) -> dict:
        """
         Set manual recovery = True | False
        """
        response = self.control.SetManualRecovery(common_msgs.State(enable=enable))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def calculate_relative_pose(self, start_pos, end_pos,
                                base_type=TaskBaseType.ABSOLUTE):
        """
        Calculate relative pose
        """
        response = self.control.CalculateRelativePose(control_msgs.CalculateRelativePoseReq(
            start_pos=list(start_pos),
            end_pos=list(end_pos),
            base_type=base_type
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    def calculate_current_pose_rel(self, current_pos, relative_pos,
                                   base_type=TaskBaseType.ABSOLUTE):
        """
        Calculate current pos rel
        """
        response = self.control.CalculateCurrentPoseRel(control_msgs.CalculateCurrentPoseRelReq(
            current_pos=list(current_pos),
            relative_pos=list(relative_pos),
            base_type=base_type
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    ############################
    # Program control
    ############################
    def play_program(self, prog_name: str = '', prog_idx: int = -1):
        """
         Play program
        """
        response = self.control.PlayProgram(control_msgs.Program(
            prog_name=prog_name,
            prog_idx=prog_idx
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def pause_program(self):
        """
         Pause program
        """
        response = self.control.PauseProgram(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def resume_program(self):
        """
         Resume program
        """
        response = self.control.ResumeProgram(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def stop_program(self):
        """
         Stop program
        """
        response = self.control.StopProgram(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_speed_ratio(self, speed_ratio: int):
        """
        Speed Ratio
            ratio -> uint32 {0 ~ 100}
        """
        response = self.config.SetSpeedRatio(config_msgs.Ratio(
            ratio=speed_ratio
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    ############################
    # Variables
    ############################
    def get_bool_variable(self):
        """
        Bool Variables:
            [
                addr -> int32
                value -> bool
            ]
        """
        response = self.control.GetBoolVariable(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_int_variable(self):
        """
        Integer Variables:
            [
                addr -> int32
                value -> int32
            ]
        """
        response = self.control.GetIntVariable(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_float_variable(self):
        """
        Float Variables:
            [
                addr -> int32
                value -> float
            ]
        """
        response = self.control.GetFloatVariable(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_jpos_variable(self):
        """
        JPos Variables:
            [
                addr -> int32
                jpos -> float[]
            ]
        """
        response = self.control.GetJPosVariable(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)['variables']

    def get_tpos_variable(self):
        """
        TPos Variables:
            [
                addr -> int32
                tpos -> float[]
            ]
        """
        response = self.control.GetTPosVariable(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_bool_variable(self, bool_variables: list):
        """
        Bool Variables:
            [
                addr -> int32
                value -> bool
            ]
        """
        variable_list = []
        for bool_var in bool_variables:
            variable_list.append(control_msgs.BoolVariable(addr=bool_var['addr'], value=bool_var['value']))

        response = self.control.SetBoolVariable(
            control_msgs.BoolVars(variables=variable_list)
        )
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_int_variable(self, int_variables: list):
        """
        Integer Variables:
            [
                addr -> int32
                value -> int64
            ]
        """
        variable_list = []
        for int_var in int_variables:
            variable_list.append(control_msgs.IntVariable(addr=int_var['addr'], value=int_var['value']))

        response = self.control.SetIntVariable(
            control_msgs.IntVars(variables=variable_list)
        )
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_float_variable(self, float_variables: list):
        """
        Float Variables:
            [
                addr -> int32
                value -> float
            ]
        """
        variable_list = []
        for float_var in float_variables:
            variable_list.append(control_msgs.FloatVariable(addr=float_var['addr'], value=float_var['value']))

        response = self.control.SetFloatVariable(
            control_msgs.FloatVars(variables=variable_list)
        )
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_jpos_variable(self, jpos_variables: list):
        """
        JPos Variables:
            [
                addr -> int32
                jpos -> float[]
            ]
        """
        variable_list = []
        for jpos in jpos_variables:
            variable_list.append(control_msgs.JPosVariable(addr=jpos['addr'], jpos=jpos['jpos']))

        response = self.control.SetJPosVariable(
            control_msgs.JPosVars(variables=variable_list)
        )
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_tpos_variable(self, tpos_variables: list):
        """
        TPos Variables:
            [
                addr -> int32
                tpos -> float[]
            ]
        """
        variable_list = []
        for tpos in tpos_variables:
            variable_list.append(control_msgs.TPosVariable(addr=tpos['addr'], tpos=tpos['tpos']))

        response = self.control.SetTPosVariable(
            control_msgs.TPosVars(variables=variable_list)
        )
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    ############################
    # Config
    ############################
    def get_home_pos(self):
        """
        Joint Home Position
            jpos -> double[]
        """
        response = self.config.GetHomePosition(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_home_pos(self, home_jpos: list):
        """
        Joint Home Position
            jpos -> double[]
        """
        response = self.config.SetHomePosition(config_msgs.JointPos(
            jpos=home_jpos
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    def get_ref_frame(self):
        """
        Reference frame
            fpos -> float[6]
        """

        response = self.config.GetRefFrame(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_ref_frame(self, fpos: list):
        """
        Ref Frame
            fpos -> float[6]
        """
        response = self.config.SetRefFrame(config_msgs.Frame(
            fpos=list(fpos)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_ref_frame_planar(self, fpos0: list, fpos1: list, fpos2: list):
        """
        Ref Frame
            fpos -> float[6]
        """
        response = self.config.SetRefFramePlanar(config_msgs.PlanarFrame(
            fpos0=list(fpos0), fpos1=list(fpos1), fpos2=list(fpos2)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_tool_frame(self, fpos: list):
        """
        Tool Frame
            fpos -> float[6]
        """
        response = self.config.SetToolFrame(config_msgs.Frame(
            fpos=list(fpos)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    def get_friction_comp(self):
        """
        Friction Compensation Set:
            joint_idx   -> uint32
            control_comp_enable   -> bool
            control_comp_levels   -> int32[6]
            teaching_comp_enable   -> bool
            teaching_comp_levels   -> int32[6]
        """
        response = self.config.GetFrictionComp(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_friction_comp(self, control_comp: bool, control_comp_levels: list,
                          dt_comp: bool, dt_comp_levels: list):
        """
        Friction Compensation Set:
            joint_idx   -> uint32
            control_comp_enable   -> bool
            control_comp_levels   -> int32[6]
            teaching_comp_enable   -> bool
            teaching_comp_levels   -> int32[6]
        """
        response = self.config.SetFrictionComp(config_msgs.FrictionCompSet(
            control_comp_enable=control_comp, control_comp_levels=list(control_comp_levels),
            teaching_comp_enable=dt_comp, teaching_comp_levels=list(dt_comp_levels)
        ))

        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    def get_tool_property(self):
        """
        Tool Properties:
            mass   -> float
            center_of_mass   -> float[3]
            inertia   -> float[6]
        """
        response = self.config.GetToolProperty(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_tool_property(self, mass: float, center_of_mass: list, inertia: list):
        """
        Tool Properties:
            mass   -> float
            center_of_mass   -> float[3]
            inertia   -> float[6]
        """
        response = self.config.SetToolProperty(config_msgs.ToolProperties(
            mass=mass, center_of_mass=list(center_of_mass), inertia=list(inertia)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_coll_sens_level(self):
        """
        Collision Sensitivity Level:
            level -> uint32
        """
        response = self.config.GetCollSensLevel(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_coll_sens_level(self, level: int):
        """
        Collision Sensitivity Level:
            level -> uint32
        """
        response = self.config.SetCollSensLevel(config_msgs.CollisionSensLevel(
            level=level
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_coll_sens_param(self):
        """
        Collision Params:
            j_torque_bases                  -> double[6]
            j_torque_tangents               -> double[6]
            t_torque_bases                  -> double[6]
            t_torque_tangents               -> double[6]
            error_bases                     -> double[6]
            error_tangents                  -> double[6]
            t_constvel_torque_bases         -> double[6]
            t_constvel_torque_tangents      -> double[6]
            t_conveyor_torque_bases         -> double[6]
            t_conveyor_torque_tangents      -> double[6]
        """
        response = self.config.GetCollSensParam(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_coll_sens_param(self, j_torque_bases, j_torque_tangents,
                            t_torque_bases, t_torque_tangents,
                            t_constvel_torque_bases, t_constvel_torque_tangents,
                            t_conveyor_torque_bases, t_conveyor_torque_tangents,
                            error_bases, error_tangents):
        """
        Collision Params:
            j_torque_bases                  -> double[6]
            j_torque_tangents               -> double[6]
            t_torque_bases                  -> double[6]
            t_torque_tangents               -> double[6]
            error_bases                     -> double[6]
            error_tangents                  -> double[6]
            t_constvel_torque_bases         -> double[6]
            t_constvel_torque_tangents      -> double[6]
            t_conveyor_torque_bases         -> double[6]
            t_conveyor_torque_tangents      -> double[6]
        """
        response = self.config.SetCollSensParam(config_msgs.CollisionThresholds(
            j_torque_bases=list(j_torque_bases), j_torque_tangents=list(j_torque_tangents),
            t_torque_bases=list(t_torque_bases), t_torque_tangents=list(t_torque_tangents),
            error_bases=list(error_bases), error_tangents=list(error_tangents),
            t_constvel_torque_bases=list(t_constvel_torque_bases),
            t_constvel_torque_tangents=list(t_constvel_torque_tangents),
            t_conveyor_torque_bases=list(t_conveyor_torque_bases),
            t_conveyor_torque_tangents=list(t_conveyor_torque_tangents)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_coll_policy(self):
        """
        Collision Policy:
            policy -> uint32
            sleep_time -> float
            gravity_time -> float
        """
        response = self.config.GetCollPolicy(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_coll_policy(self, policy=CollisionPolicyType.NONE,
                        sleep_time=0, gravity_time=0.1):
        """
        Collision Policies:
            policy -> uint32
            sleep_time -> float
            gravity_time -> float
        """
        CollisionPolicyType.NONE
        response = self.config.SetCollPolicy(config_msgs.CollisionPolicy(
            policy=policy, sleep_time=sleep_time, gravity_time=gravity_time
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_safety_limits(self):
        """
        Safety Limits:
            power_limit             -> float
            power_limit_ratio       -> float
            tcp_force_limit         -> float
            tcp_force_limit_ratio   -> float
            tcp_speed_limit         -> float
            tcp_speed_limit_ratio   -> float
            joint_limits   -> float[]
        """
        response = self.config.GetSafetyLimits(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_safety_limits(self, power_limit: float, power_limit_ratio: float,
                          tcp_force_limit: float, tcp_force_limit_ratio: float,
                          tcp_speed_limit: float, tcp_speed_limit_ratio: float):
        # joint_limits: list):
        """
        Safety Limits:
            power_limit             -> float
            power_limit_ratio       -> float
            tcp_force_limit         -> float
            tcp_force_limit_ratio   -> float
            tcp_speed_limit         -> float
            tcp_speed_limit_ratio   -> float
            # joint_limits   -> float[]
        """
        response = self.config.SetSafetyLimits(config_msgs.SafetyLimits(
            power_limit=power_limit, power_limit_ratio=power_limit_ratio,
            tcp_force_limit=tcp_force_limit, tcp_force_limit_ratio=tcp_force_limit_ratio,
            tcp_speed_limit=tcp_speed_limit, tcp_speed_limit_ratio=tcp_speed_limit_ratio  # ,
            # joint_limits=list(joint_limits)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    ############################
    # IndySDK related
    ############################
    def activate_sdk(self, license_key, expire_date):
        """
        license_key: license key issued by Neuromeka
        expire_date: expire date for the license, format YYYY-MM-DD
        SDKLicenseResp:
            activated -> bool, True if activated
            response (code, msg)
                - 0, 'Activated'                -> SDK Activated
                - 1, 'Invalid'                  -> Wrong key or expire date
                - 2, 'No Internet Connection'   -> Need Internet for License Verification
                - 3, 'Expired'                  -> License Expired
                - 4, 'HW_FAILURE'               -> Failed acquire HW ID to verify license
        """
        response = self.control.ActivateIndySDK(
            control_msgs.SDKLicenseInfo(license_key=license_key, expire_date=expire_date))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_custom_control_mode(self, mode):
        """
        mode:
        - False (0): IndyFramework's default controller is used
        - True (1): IndySDK's component is used
        """
        response = self.control.SetCustomControlMode(common_msgs.IntMode(mode=mode))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_custom_control_mode(self):
        """

        """
        response = self.control.GetCustomControlMode(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    ############################
    # Uility functions
    ############################
    def start_log(self):
        """
        Start realtime data logging
        """
        int_vars_to_set = [{"addr": 300, "value": 1}]
        self.set_int_variable(int_vars_to_set)

    def end_log(self):
        """
        Finish realtime data logging and save the realtime data in STEP
        saved path:
            /home/user/release/IndyDeployments/RTlog/RTLog.csv
        """
        int_vars_to_set = [{"addr": 300, "value": 2}]
        self.set_int_variable(int_vars_to_set)

    ############################
    # NOT Yet Released
    ############################
    def wait_time(self, time: float,
                  set_do_signal_list=None, set_end_do_signal_list=None,
                  set_ao_signal_list=None, set_end_ao_signal_list=None):
        """
         Wait time [s]
        """
        response = self.control.WaitTime(control_msgs.WaitTimeReq(
            time=time,
            set_do_list=self.__to_digital_request_list__(set_do_signal_list),
            set_end_do_list=self.__to_digital_request_list__(set_end_do_signal_list),
            set_ao_list=self.__to_analog_request_list__(set_ao_signal_list),
            set_end_ao_list=self.__to_analog_request_list__(set_end_ao_signal_list)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def wait_progress(self, progress: int,
                      set_do_signal_list=None, set_end_do_signal_list=None,
                      set_ao_signal_list=None, set_end_ao_signal_list=None):
        """
         Wait progress [s]
        """
        response = self.control.WaitProgress(control_msgs.WaitProgressReq(
            progress=progress,
            set_do_list=self.__to_digital_request_list__(set_do_signal_list),
            set_end_do_list=self.__to_digital_request_list__(set_end_do_signal_list),
            set_ao_list=self.__to_analog_request_list__(set_ao_signal_list),
            set_end_ao_list=self.__to_analog_request_list__(set_end_ao_signal_list)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def wait_traj(self, traj_condition,
                  set_do_signal_list=None, set_end_do_signal_list=None,
                  set_ao_signal_list=None, set_end_ao_signal_list=None):
        """
         Wait trajectory
        """
        response = self.control.WaitTraj(control_msgs.WaitTrajReq(
            traj_condition=traj_condition,
            set_do_list=self.__to_digital_request_list__(set_do_signal_list),
            set_end_do_list=self.__to_digital_request_list__(set_end_do_signal_list),
            set_ao_list=self.__to_analog_request_list__(set_ao_signal_list),
            set_end_ao_list=self.__to_analog_request_list__(set_end_ao_signal_list)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def wait_radius(self, radius: int,
                    set_do_signal_list=None, set_end_do_signal_list=None,
                    set_ao_signal_list=None, set_end_ao_signal_list=None):
        """
         Wait radius [mm]
        """
        response = self.control.WaitRadius(control_msgs.WaitRadiusReq(
            radius=radius,
            set_do_list=self.__to_digital_request_list__(set_do_signal_list),
            set_end_do_list=self.__to_digital_request_list__(set_end_do_signal_list),
            set_ao_list=self.__to_analog_request_list__(set_ao_signal_list),
            set_end_ao_list=self.__to_analog_request_list__(set_end_ao_signal_list)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_do_config_list(self, do_config_list: dict):
        """
        DO Configuration List
            {
                'do_configs': [
                    {
                        'state_code': 2,
                        'state_name': "name",
                        'onSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}],
                        'offSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}]
                    }
                ]
            }
        """
        do_list_request = config_msgs.DOConfigList()
        json_format.ParseDict(do_config_list, do_list_request)

        response = self.config.SetDOConfigList(do_list_request)

        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_do_config_list(self):
        """
        DO Configuration List
            {
                'do_configs': [
                    {
                        'state_code': 2,
                        'state_name': "name",
                        'onSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}],
                        'offSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}]
                    }
                ]
            }
        """
        response = self.config.GetDOConfigList(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def move_recover_joint(self, jtarget,
                           base_type=JointBaseType.ABSOLUTE) -> dict:
        """
         Move recover joint
         jtarget = [deg, deg, deg, deg, deg, deg]
        """
        response = self.control.MoveRecoverJoint(
            control_msgs.TargetJ(j_target=list(jtarget), base_type=base_type)
        )
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_control_info(self):
        response = self.control.GetControlInfo(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    def check_aproach_retract_valid(self, tpos, init_jpos, pre_tpos, post_tpos):
        """
        Check aproach retract valid
        """
        response = self.control.CheckAproachRetractValid(control_msgs.CheckAproachRetractValidReq(
            tpos=list(tpos),
            init_jpos=list(init_jpos),
            pre_tpos=list(pre_tpos),
            post_tpos=list(post_tpos)
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_pallet_point_list(self, tpos, jpos, pre_tpos, post_tpos, pallet_pattern, width, height):
        """
        Get pallet point list
        """
        response = self.control.GetPalletPointList(control_msgs.GetPalletPointListReq(
            tpos=list(tpos),
            jpos=list(jpos),
            pre_tpos=list(pre_tpos),
            post_tpos=list(post_tpos),
            pallet_pattern=pallet_pattern,
            width=width,
            height=height
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)


    def play_tuning_program(self, prog_name: str = '', prog_idx: int = -1,
                            tuning_space=common_msgs.TUNE_ALL, precision=common_msgs.HIGH_PRECISION,
                            vel_level_max=9):
        """
        Play tuning program
        """
        tuning_prog_dict = dict(
            program=dict(
                prog_name=prog_name,
                prog_idx=prog_idx),
            tuning_space=tuning_space,
            precision=precision,
            vel_level_max=vel_level_max
        )
        tuning_req = control_msgs.TuningProgram()

        # json_format.ParseDict(tuning_prog_dict, tuning_req)
        ParseDict(tuning_prog_dict, tuning_req)
        response = self.control.PlayTuningProgram(tuning_req)
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_di_config_list(self, di_config_list: dict):
        """
        DI Configuration List
            {
                'di_configs': [
                    {
                        'function_code': 2,
                        'function_name': "name",
                        'triggerSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}]
                        'successSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}]
                        'failureSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}]
                    }
                ]
            }
        """
        di_list_request = config_msgs.DIConfigList()
        # json_format.ParseDict(di_config_list, di_list_request)
        ParseDict(di_config_list, di_list_request)
        response = self.config.SetDIConfigList(di_list_request)

        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_di_config_list(self):
        """
        DI Configuration List
            {
                'di_configs': [
                    {
                        'function_code': 2,
                        'function_name': "name",
                        'triggerSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}],
                        'successSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}],
                        'failureSignals': [{'address': 1, 'state': 1}, {'address': 2, 'state': 0}]
                    }
                ]
            }
        """
        response = self.config.GetDIConfigList(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)



    def set_auto_servo_Off(self, enable: bool, time: float):
        """
        Auto Servo-Off Config
            enable -> bool
            time -> float
        """
        response = self.config.SetAutoServoOff(config_msgs.AutoServoOffConfig(
            enable=enable, time=time
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_auto_servo_off(self):
        """
        Auto Servo-Off Config
            enable -> bool
            time -> float
        """
        response = self.config.GetAutoServoOff(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def set_safety_stop_config(self, jpos_limit_stop_cat=StopCategory.CAT0,
                               jvel_limit_stop_cat=StopCategory.CAT0,
                               jtau_limit_stop_cat=StopCategory.CAT0,
                               tvel_limit_stop_cat=StopCategory.CAT0,
                               tforce_limit_stop_cat=StopCategory.CAT0,
                               power_limit_stop_cat=StopCategory.CAT0):
        """
        Safety Stop Category:
            jpos_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            jvel_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            jtau_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            tvel_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            tforce_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            power_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
        """
        response = self.config.SetSafetyStopConfig(config_msgs.SafetyStopConfig(
            joint_position_limit_stop_cat=jpos_limit_stop_cat,
            joint_speed_limit_stop_cat=jvel_limit_stop_cat,
            joint_torque_limit_stop_cat=jtau_limit_stop_cat,
            tcp_speed_limit_stop_cat=tvel_limit_stop_cat,
            tcp_force_limit_stop_cat=tforce_limit_stop_cat,
            power_limit_stop_cat=power_limit_stop_cat
        ))
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_safety_stop_config(self):
        """
        Safety Stop Category:
            joint_position_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            joint_speed_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            joint_torque_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            tcp_speed_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            tcp_force_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
            power_limit_stop_cat = IMMEDIATE_BRAKE(0) | SMOOTH_BRAKE(1) | SMOOTH_ONLY(2)
        """
        response = self.config.GetSafetyStopConfig(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)

    def get_io_data(self):
        """
        IO Data:
            di   -> DigitalSignal[]
            do   -> DigitalSignal[]
            ai  -> AnalogSignal[]
            ao  -> AnalogSignal[]
            end_di  -> EndtoolSignal[]
            end_do  -> EndtoolSignal[]
            end_ai  -> AnalogSignal[]
            end_ao  -> AnalogSignal[]
            response  -> Response
        """
        response = self.rtde.GetIOData(common_msgs.Empty())
        return json_format.MessageToDict(response,
                                         including_default_value_fields=True,
                                         preserving_proto_field_name=True,
                                         use_integers_for_enums=True)